import abc
from typing import Generic, TypeVar

Input = TypeVar("Input")
Output = TypeVar("Output")


class Model(abc.ABC, Generic[Input, Output]):
    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def model(self, data: Input) -> Output:
        pass



from typing import Any, Dict, List, Sequence, Union

import numpy as np
import torch

from embeddings.evaluator.metrics_evaluator import MetricsEvaluator
from embeddings.metric.hugging_face_metric import HuggingFaceMetric
from embeddings.metric.metric import Metric


class TextClassificationEvaluator(MetricsEvaluator):
    @property
    def metrics(
        self,
    ) -> Sequence[Metric[Union[List[Any], np.ndarray, torch.Tensor], Dict[Any, Any]]]:
        return [
            HuggingFaceMetric("accuracy"),
            HuggingFaceMetric("f1", compute_kwargs={"average": "macro"}),
            HuggingFaceMetric("recall", compute_kwargs={"average": "macro"}),
            HuggingFaceMetric("precision", compute_kwargs={"average": "macro"}),
        ]



# from pathlib import Path
# from typing import Optional

# import flair
# import torch
# import typer
# from flair.embeddings import TransformerWordEmbeddings

# import experimental.datasets.utils.misc
# from embeddings.defaults import RESULTS_PATH

# app = typer.Typer()


# def setup(device: str) -> None:
#     flair.device = torch.device(device)  # type:ignore
#     flair.set_seed(441)


# def evaluate_sequence_tagging(
#     embedding: str = typer.Option(..., help="Huggingface embedding model name"),
#     dataset_name: str = typer.Option(..., help="Dataset name"),
#     dataset_url: Optional[str] = typer.Option(None, help="Dataset download url"),
#     root: str = typer.Option(
#         RESULTS_PATH.joinpath("tagging"),
#         help="Root directory for output files",
#     ),
#     device: str = typer.Option("cpu", help="Device for Flair e.g. cpu"),
# ) -> None:
#     setup(device)

#     out_dir = Path(root, embedding, dataset_name)
#     out_dir.mkdir(parents=True, exist_ok=True)

#     ds_args = {"url": dataset_url} if dataset_url else {}
#     dataset_cls = experimental.datasets.utils.misc.get_dataset_cls(dataset_name)
#     dataset = dataset_cls(**ds_args)
#     corpus = dataset.to_flair_column_corpus()
#     tag_dictionary = corpus.make_tag_dictionary(tag_type="tag")

#     test_sentences = corpus.test.sentences.copy()

#     print("Loading embeddings...")
#     embeddings = TransformerWordEmbeddings(embedding)

#     print("Training model...")
#     tagger = FlairSequenceTagger(
#         embeddings=embeddings,
#         hidden_dim=256,
#         tag_dictionary=tag_dictionary,
#         output_path=str(out_dir),
#     )

#     training_log = tagger.fit(corpus)
#     evaluation_log = tagger.evaluate(test_sentences)
#     with out_dir.joinpath("results.pkl").open(mode="wb") as f:
#         pickle.dump(obj={"training_log": training_log, "evaluation_log": evaluation_log}, file=f)

#     print("Done!", training_log)


# typer.run(evaluate_sequence_tagging)



from typing import Any, Dict, List, Sequence, Union

import numpy as np
import torch

from embeddings.evaluator.metrics_evaluator import MetricsEvaluator
from embeddings.metric.hugging_face_metric import HuggingFaceMetric
from embeddings.metric.metric import Metric
from embeddings.metric.pos_metric import POSTaggingSeqevalMetric


class SequenceLabelingEvaluator(MetricsEvaluator):
    @property
    def metrics(
        self,
    ) -> Sequence[Metric[Union[List[Any], np.ndarray, torch.Tensor], Dict[Any, Any]]]:
        return [
            HuggingFaceMetric("seqeval"),
        ]


class POSTaggingEvaluator(MetricsEvaluator):
    @property
    def metrics(
        self,
    ) -> Sequence[POSTaggingSeqevalMetric]:
        return [
            POSTaggingSeqevalMetric(),
        ]



from pathlib import Path
from typing import Dict

import datasets
from flair.datasets import ColumnDataset

from embeddings.transformation.flair_transformation.corpus_transformation import (
    CorpusTransformation,
)


class ColumnCorpusTransformation(CorpusTransformation):
    def _preprocess_subset(
        self, hf_datadict: datasets.DatasetDict, subset_name: str, output_path: Path
    ) -> ColumnDataset:
        tag_map = hf_datadict[subset_name].features[self.target_column_name].feature.names

        hf_datadict[subset_name] = hf_datadict[subset_name].map(
            lambda row: {
                self.target_column_name: [
                    tag_map[tag_idx] for tag_idx in row[self.target_column_name]
                ]
            },
        )

        column_format = self._save_to_conll(
            hf_datadict[subset_name], output_path.joinpath(f"{subset_name}.csv")
        )

        return ColumnDataset(output_path.joinpath(f"{subset_name}.csv"), column_format)

    def _check_task(self, dataset: datasets.Dataset) -> None:
        if not self._is_input_sequence_of_strings(dataset):
            raise ValueError(
                f"Type of input column is not '{datasets.Sequence.__name__}' of"
                f" '{datasets.Value.__name__}' with strings."
            )

        if not self._is_target_sequence_of_labels(dataset):
            raise ValueError(
                f"Type of target column is not '{datasets.Sequence.__name__}' of"
                f" '{datasets.ClassLabel.__name__}'."
            )

    def _is_input_sequence_of_strings(self, dataset: datasets.Dataset) -> bool:
        if not isinstance(dataset.features[self.input_column_name], datasets.Sequence):
            return False
        elif not isinstance(dataset.features[self.input_column_name].feature, datasets.Value):
            return False
        elif dataset.features[self.input_column_name].feature.dtype != "string":
            return False
        else:
            return True

    def _is_target_sequence_of_labels(self, dataset: datasets.Dataset) -> bool:
        if not isinstance(dataset.features[self.target_column_name], datasets.Sequence):
            return False
        elif not isinstance(dataset.features[self.target_column_name].feature, datasets.ClassLabel):
            return False
        else:
            return True

    def _save_to_conll(self, dataset: datasets.Dataset, output_path: Path) -> Dict[int, str]:
        with open(output_path, "w", encoding="utf-8") as f:
            for tokens, tags in zip(
                dataset[self.input_column_name], dataset[self.target_column_name]
            ):
                assert len(tokens) == len(tags)
                for token, tag in zip(tokens, tags):
                    f.write(f"{token}\t{tag}\n")
                f.write("\n")

        column_format = {0: "text", 1: "tag"}
        return column_format



import datasets

from embeddings.data.data_loader import DataLoader
from embeddings.data.dataset import Dataset
from embeddings.data.hugging_face_dataset import HuggingFaceDataset


class HuggingFaceDataLoader(DataLoader[str, datasets.DatasetDict]):
    def load(self, dataset: Dataset[str]) -> datasets.DatasetDict:
        if isinstance(dataset, HuggingFaceDataset):
            return datasets.load_dataset(dataset.dataset, **dataset.load_dataset_kwargs)
        else:
            raise ValueError("This DataLoader should be used with HuggingFaceDataset only.")



import abc
import pprint
import tempfile
from itertools import groupby
from operator import itemgetter
from pathlib import Path
from typing import Dict

import datasets
from flair.data import Corpus, FlairDataset

from embeddings.defaults import DATASET_PATH
from embeddings.transformation.transformation import Transformation
from embeddings.utils.loggers import get_logger

_logger = get_logger(__name__)


class CorpusTransformation(Transformation[datasets.DatasetDict, Corpus]):
    HUGGING_FACE_SUBSETS = ["train", "validation", "test"]

    def __init__(
        self,
        input_column_name: str,
        target_column_name: str,
        datasets_path: Path = DATASET_PATH,
    ):
        super().__init__()

        self.datasets_cache_path = datasets_path
        self.input_column_name = input_column_name
        self.target_column_name = target_column_name

    def transform(self, data: datasets.DatasetDict) -> Corpus:
        with tempfile.TemporaryDirectory() as tmp_dir_path:
            output_path = Path(tmp_dir_path)
            flair_datasets = self._preprocess(data, output_path)
        return self._to_flair_corpus(flair_datasets)

    @staticmethod
    def _to_flair_corpus(flair_datasets: Dict[str, FlairDataset]) -> Corpus:
        if not flair_datasets["train"]:
            raise ValueError("Hugging Face dataset does not contain TRAIN subset.")

        return Corpus(
            train=flair_datasets["train"],
            dev=flair_datasets["validation"],
            test=flair_datasets["test"],
        )

    def _preprocess(
        self, hf_datadict: datasets.DatasetDict, output_path: Path
    ) -> Dict[str, FlairDataset]:
        flair_datasets = {}
        self._log_info(hf_datadict)

        for subset_name in hf_datadict.keys():
            self._check_compatibility(hf_datadict[subset_name])

        for subset_name in CorpusTransformation.HUGGING_FACE_SUBSETS:
            if subset_name in hf_datadict.keys():
                flair_datasets[subset_name] = self._preprocess_subset(
                    hf_datadict, subset_name, output_path
                )
            else:
                flair_datasets[subset_name] = None
        return flair_datasets

    @abc.abstractmethod
    def _preprocess_subset(
        self, hf_datadict: datasets.DatasetDict, subset_name: str, output_path: Path
    ) -> FlairDataset:
        pass

    @staticmethod
    def _log_info(hf_datadict: datasets.DatasetDict) -> None:
        subsets_info = {
            subset: pprint.pformat(hf_datadict[subset].info.__dict__)
            for subset in hf_datadict.keys()
        }
        for k, v in groupby(subsets_info.items(), itemgetter(1)):
            _logger.info(f"Info of {list(map(itemgetter(0), v))}:\n{k}")
        _logger.info(f"Schemas:\t{hf_datadict}")

    def _check_compatibility(self, dataset: datasets.Dataset) -> None:
        self._check_column_in_dataset(dataset, self.input_column_name)
        self._check_column_in_dataset(dataset, self.target_column_name)
        self._check_task(dataset)

    @staticmethod
    def _check_column_in_dataset(dataset: datasets.Dataset, column_name: str) -> None:
        if column_name not in dataset.features:
            raise KeyError(f"Column '{column_name}' not found in features.")

    @abc.abstractmethod
    def _check_task(self, dataset: datasets.Dataset) -> None:
        """Checking if dataset is compatible with the given task."""
        pass



from transformers import AutoConfig, AutoModel, AutoTokenizer

MODEL = "clarin-pl/long-former-polish"

config = AutoConfig.from_pretrained(MODEL)
tokenizer = AutoTokenizer.from_pretrained(MODEL)
model = AutoModel.from_pretrained(MODEL)

encoded_input = tokenizer.encode("Zażółć gęślą jaźń", return_tensors="pt")
outputs = model(encoded_input)



from typing import Any, Dict, List, Optional

import numpy as np
from flair.data import Corpus, Dictionary, Sentence
from flair.models import TextClassifier

from embeddings.data.io import T_path
from embeddings.embedding.flair_embedding import FlairEmbedding
from embeddings.task.flair_task.flair_task import FlairTask


class TextClassification(FlairTask):
    def __init__(
        self,
        output_path: T_path,
        task_model_kwargs: Optional[Dict[str, Any]] = None,
        task_train_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(output_path, task_train_kwargs)
        self.task_model_kwargs = task_model_kwargs if task_model_kwargs else {}

    def build_task_model(self, embedding: FlairEmbedding, y_dictionary: Dictionary) -> None:
        self.model = TextClassifier(
            document_embeddings=embedding.model,
            label_dictionary=y_dictionary,
            **self.task_model_kwargs
        )

    @property
    def y_type(self) -> Any:
        if self.model:
            return self.model.label_type
        else:
            raise self.MODEL_UNDEFINED_EXCEPTION

    @staticmethod
    def make_y_dictionary(data: Corpus, y_type: Optional[str] = None) -> Dictionary:
        return data.make_label_dictionary(y_type)

    @property
    def y_dictionary(self) -> Dictionary:
        if self.model:
            return self.model.label_dictionary
        else:
            raise self.MODEL_UNDEFINED_EXCEPTION

    @staticmethod
    def get_y(data: List[Sentence], y_type: str, y_dictionary: Dictionary) -> np.ndarray:
        labels = [sentence.get_labels(y_type) for sentence in data]
        labels = [[label.value for label in sentence] for sentence in labels]
        labels = [y_dictionary.get_idx_for_item(labels[0]) for labels in labels]
        return np.array(labels)

    @staticmethod
    def remove_labels_from_data(data: List[Sentence], y_type: str) -> None:
        for sentence in data:
            sentence.remove_labels(y_type)



import abc
from abc import ABC
from typing import Generic, TypeVar

Output = TypeVar("Output")


class Pipeline(ABC, Generic[Output]):
    @abc.abstractmethod
    def run(self) -> Output:
        pass



import abc
from abc import ABC
from typing import Any, Generic, TypeVar

Input = TypeVar("Input")
Output = TypeVar("Output")


class Metric(ABC, Generic[Input, Output]):
    def __init__(self, name: str) -> None:
        self._name = name

    def __str__(self) -> str:
        return self._name

    @abc.abstractmethod
    def compute(self, y_true: Input, y_pred: Input, **kwargs: Any) -> Output:
        pass
